import numpy

from gdshoplib.apps.crm.orders import Order
from gdshoplib.services.notion.notion import Notion


class Statistic:
    def avg_price(cls, filter_type="complited"):
        params = Order.list_filter(filter_type=filter_type)
        notion = Notion(caching=True)
        return numpy.average(
            [order.price for order in Order.query(notion=notion, params=params)]
        )

    def percentile(self, filter_type="complited"):
        params = Order.list_filter(filter_type=filter_type)
        notion = Notion(caching=True)
        return numpy.percentile(
            [order.price for order in Order.query(notion=notion, params=params)], 5
        )



import datetime

from gdshoplib.services.vk.vk import VK


class VKStats:
    def __init__(self, manager=None):
        self.manager = manager or VK()

    def members(self):
        # Возвращает список участников сообщества.
        result = self.manager.request(
            "groups.getMembers",
            params={
                "group_id": self.manager.settings.VK_GROUP_ID,
                "fields": [
                    "bdate",
                    "can_post",
                    "can_see_all_posts",
                    "can_see_audio",
                    "can_write_private_message",
                    "city",
                    "common_count",
                    "connections",
                    "contacts",
                    "country",
                    "domain",
                    "education",
                    "has_mobile",
                    "last_seen",
                    "lists",
                    "online",
                    "online_mobile",
                    "photo_100",
                    "photo_200",
                    "photo_200_orig",
                    "photo_400_orig",
                    "photo_50",
                    "photo_max",
                    "photo_max_orig",
                    "relation",
                    "relatives",
                    "schools",
                    "sex",
                    "site",
                    "status",
                    "universities",
                ],
            },
        )
        return result

    def products(self):
        # Возвращает информацию о товаре
        result = self.manager.request(
            "market.get",
            params={"owner_id": f"-{self.manager.settings.VK_GROUP_ID}", "extended": 1},
        )
        return result

    def community(
        self,
        timestamp_from: datetime.datetime = None,
        timestamp_to: datetime.datetime = None,
    ):
        # Возвращает статистику сообщества или приложения.
        result = self.manager.request(
            "stats.get",
            params={
                "group_id": self.manager.settings.VK_GROUP_ID,
                "timestamp_from": (
                    timestamp_from
                    or datetime.datetime.now() - datetime.timedelta(days=1)
                ).timestamp(),
                "timestamp_to": (timestamp_to or datetime.datetime.now()).timestamp(),
                "interval": "day",
                "extended": 1,
            },
        )
        return result



from gdshoplib.services.notion.notion import Notion


class BasePage:
    def __init__(self, id, *, notion=None, parent=None):
        self.properties = None
        self.id = id
        self.notion = notion or Notion()
        self.parent = parent
        self.history = {}
        self.change_log = {}
        self.initialize()

    def initialize(self):
        self.page = self.get_source()

    def get_source(self):
        return self.notion.get_page(self.id)

    def __str__(self) -> str:
        return f"{self.__class__}: {self.id}"

    def __repr__(self) -> str:
        return f"{self.__class__}: {self.id}"

    def __getitem__(self, key):
        try:
            return super(BasePage, self).__getattribute__(key)
        except AttributeError:
            return self.__getattr__(key)

    def __getattr__(self, name: str):
        if self.page and name in self.page.keys():
            return self.page[name]

        if self.properties:
            return self.properties[name]

        raise AttributeError

    def __enter__(self):
        self.notion.cache.set_update_start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if not exc_type:
            self.set_update_time()
        self.notion.cache.clean_update()

    def get_update_time(self):
        return self.notion.cache.get_update_time(self.id, type=self.__class__.__name__)

    def set_update_time(self):
        return self.notion.cache.set_update_time(self.id, type=self.__class__.__name__)

    def edited_filter(self):
        update_time = self.get_update_time()
        if not update_time:
            return None
        return {
            "filter": {
                "timestamp": "last_edited_time",
                "last_edited_time": {"after": update_time},
            },
            "sorts": [
                {"direction": "ascending", "timestamp": "last_edited_time"},
            ],
        }



import time
from datetime import datetime
from multiprocessing import Pool
from typing import List, Optional

import typer

from gdshoplib.apps.products.product import Product
from gdshoplib.apps.uploader.uploader import Uploader
from gdshoplib.core.ecosystem import Ecosystem
from gdshoplib.core.settings import NotionSettings
from gdshoplib.packages.cache import KeyDBCache
from gdshoplib.services.notion.database import Database
from gdshoplib.services.notion.notion import Notion

app = typer.Typer()


@app.command()
def clean(id: Optional[str] = typer.Option(None)):
    # TODO: сделать удаление по ID
    KeyDBCache().clean(r"[blocks|pages|databases]*")


@app.command()
def listen(topic: List[str]):
    for message in KeyDBCache().subscribe(topic):
        print(message)


@app.command()
def upload(page_id: Optional[str] = typer.Option(None), infinity: bool = False):
    uploader = Uploader(NotionSettings().KAFKA_TOPIC)
    notion = Notion()

    last_edited = None
    new_time = None
    while True:
        _itarator = notion.search()

        if page_id:
            _itarator = [Notion().get_page(page_id)]

        for page in _itarator:
            page_last_edited = datetime.strptime(
                page["last_edited_time"].split(".")[0], "%Y-%m-%dT%H:%M:%S"
            )
            if not last_edited:
                last_edited = page_last_edited

            if infinity and page_last_edited <= last_edited:
                break

            uploader.send(page)
            for block in notion.get_blocks_iterator(page["id"]):
                uploader.send(block, parent=page["id"])

            if page_last_edited > last_edited:
                new_time = page_last_edited

        last_edited = new_time

        if not infinity:
            break


@app.command()
def scan():
    result = {"all": []}

    for page in Notion().all():
        if not result.get(page["object"]):
            result[page["object"]] = []

        result[page["object"]].append(page["id"])
        result["all"].append(page["id"])

    Ecosystem().send_message(
        NotionSettings().KAFKA_TOPIC, data=result, message_type="scan"
    )


@app.command()
def warm(
    only_exists: bool = typer.Option(False),
    single: bool = typer.Option(False),
    only_edited: bool = typer.Option(True),
    sku: Optional[str] = typer.Option(None),
    loop_iteration: Optional[int] = typer.Option(None),
):
    while True:
        if sku:
            cache_warm_func(Product.get(sku).id)
            return

        if single:
            with Database(
                Product.SETTINGS.PRODUCT_DB, notion=Notion(caching=True)
            ) as database:
                params = {}
                if only_edited and database.get_update_time():
                    print(f"Фильтрация от даты: {database.get_update_time()}")
                    params = database.edited_filter()

                for product in database.pages(params=params):
                    skipped = False
                    if only_exists:
                        if KeyDBCache().exists(product["id"]):
                            print(f"{product['id']}: SKIPPED")
                            skipped = True

                    if not skipped:
                        cache_warm_func(product["id"])
        else:
            with Pool(3) as p:
                with Database(
                    Product.SETTINGS.PRODUCT_DB, notion=Notion(caching=True)
                ) as database:
                    params = {}
                    if only_edited and database.get_update_time():
                        print(f"Фильтрация от даты: {database.get_update_time()}")
                        params = database.edited_filter()

                    for product in database.pages(params=params):
                        skipped = False
                        if only_exists:
                            if KeyDBCache().exists(product["id"]):
                                print(f"{product['id']}: SKIPPED")
                                skipped = True

                        if not skipped:
                            p.apply_async(cache_warm_func, (product["id"],))
                p.close()
                p.join()

        if loop_iteration:
            print("-" * 20)
            time.sleep(loop_iteration)
        else:
            break


@app.command()
def count():
    print(KeyDBCache().count())


@app.command()
def check(single: bool = typer.Option(False)):
    if single:
        for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
            cache_check_action(product["id"])
    else:
        with Pool(3) as p:
            for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                p.apply_async(cache_check_action, (product["id"],))
            p.close()
            p.join()


#####


def cache_check_action(id):
    for block in Notion().get_blocks(id):
        exists = KeyDBCache().exists(block["id"])
        print(f"{block['id']}: {exists}")


def cache_warm_func(id):
    product = Product(id)
    try:
        product.warm()
    except AttributeError:
        print(f"!= Пропущено {product.sku}: {product.last_edited_time}")
    else:
        print(f"{product.sku}: {product.last_edited_time}")


# if __name__ == "__main__":
#     app()



import datetime
import functools

from dateutil.parser import parse

from gdshoplib.core.settings import PriceSettins

price_settings = PriceSettins()


class Price:
    def __init__(self, product):
        self.product = product

    def get_score(self):
        # Получить текущий % наценки
        return (self.allowance_score or 0) + -(self.time_discount or 0)

    def get_kit_price(self, base_price="now"):
        if not self.product.kit:
            return None

        result = self[base_price] if self.product.quantity else 0
        for product in self.product.kit:
            if product.quantity:
                result += product.price[base_price]
        return result

    @property
    def allowance_score(self):
        # Наценка категорий
        categories_score = 0
        for category in self.product.categories:
            categories_score += category.price_coefficient

        # Наценка бренда
        brand_score = self.product.brand.price_coefficient if self.product.brand else 0
        return sum([self.product.price_coefficient, brand_score, categories_score])

    @property
    def current_discount(self):
        # Получить текущую скидку
        return 100 - round(self.now / self.profit * 100)

    @property
    def time_discount(self):
        created_time = self.product.discount_from_date or self.product.created_time
        created_time = (
            parse(created_time).date()
            if isinstance(created_time, str)
            else created_time
        )
        created_at = (datetime.date.today() - created_time).days

        if created_at > 60:
            return 15
        elif created_at > 30:
            return 10

        return 0

    def handle_ratio(*rations):
        def decor(func):
            @functools.wraps(func)
            def wrap(self, *args, **kwargs):
                ration = sum([1, *rations])
                return func(self, *args, **kwargs) * ration

            return wrap

        return decor

    def round(func):
        @functools.wraps(func)
        def wrap(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            return int(round(result, 0))

        return wrap

    def psycho_price(func):
        def wrap(self, *args, **kwargs):
            # Последняя цифра всегда 0
            # Должно оканчиваться на 50, 00, 90
            # Цифры меньше 1000 не округлять 100-ые
            # Округлять можно в пределах от 0[текущая] до (текущая - безубыточность)*0.1
            # Приоритеты округления:
            #   - Уменьшение старшего порядка от 1к
            #   - Округление до 900
            #   - Уменьшение сотой
            #   - Выбор окончания
            price = func(self, *args, **kwargs)
            lowest = price - (price - self.neitral) * 0.1
            result = price
            highest_result = price
            while result > lowest:
                if str(result)[-3::] == "900":
                    return result
                if str(result)[-2::] in ("00", "50", "90") and highest_result == price:
                    highest_result = result
                result -= 1

            return highest_result

        return wrap

    @property
    @psycho_price
    @round
    def now(self):
        if not self.product.quantity:
            return self.neitral

        discount = self.get_score()
        if discount:
            _now = self.profit + self.profit * (discount * 0.01)
            if _now < self.neitral:
                return self.neitral
            return _now

        return self.profit

    @property
    def eur(self):
        # Цена в EUR
        return self.product.price_eur

    @property
    @round
    def net(self):
        # Цена в рублях
        return self.eur * price_settings.EURO_PRICE

    @property
    @round
    @handle_ratio(price_settings.PRICE_VAT_RATIO)
    def gross(self):
        # Цена с учетом расходов и налогов на закупку
        return self.net

    @property
    @round
    @handle_ratio(price_settings.PRICE_VAT_RATIO, price_settings.PRICE_NEITRAL_RATIO)
    def neitral(self):
        # Цена безубыточности
        return self.net

    @property
    @round
    @handle_ratio(
        price_settings.PRICE_VAT_RATIO,
        price_settings.PRICE_NEITRAL_RATIO,
        price_settings.PRICE_PROFIT_RATIO,
    )
    def profit(self):
        return self.net

    def __getitem__(self, key):
        try:
            return super(Price, self).__getattribute__(key)
        except AttributeError:
            return self.__getattr__(key)



import re

from lxml import etree, objectify

from gdshoplib.apps.platforms.base import Platform
from gdshoplib.core.settings import ProductSettings
from gdshoplib.packages.feed import Feed
from gdshoplib.services.notion.database import Database


class YandexMarketManager(Platform, Feed):
    KEY = "YAM"

    def get_categories(self):
        categories = Database(ProductSettings().CATEGORY_DB).pages()
        return [(self.fetch_id(i.id), i.title) for i in categories]

    @staticmethod
    def fetch_id(id):
        return str(sum([int(i) for i in re.sub(r"\D*", "", id)]))

    @property
    def product_filter(self):
        return dict(status_description="Готово", status_publication="Публикация")

    def get_shop(self):
        shop = objectify.Element("shop")

        shop.name = self.feed_settings.SHOP_NAME
        shop.company = self.feed_settings.COMPANY_NAME
        shop.url = self.feed_settings.SHOP_URL

        currencies = objectify.Element("currencies")
        currency = etree.Element("currency")
        currency.attrib["id"] = "RUB"
        currency.attrib["rate"] = "1"
        objectify.deannotate(currency, cleanup_namespaces=True, xsi_nil=True)
        objectify.deannotate(currencies, cleanup_namespaces=True, xsi_nil=True)
        currencies.append(currency)
        shop.currencies = currencies

        categories = objectify.Element("categories")
        for ind, k in enumerate(
            ["Все товары", "Спорт и отдых", "Конный спорт"], start=1
        ):
            category = etree.Element("category")
            category.attrib["id"] = str(ind)
            if ind - 1 != 0:
                category.attrib["parentId"] = str(ind - 1)
            category.text = k
            categories.append(category)

        for _id, name in self.get_categories():
            category = etree.Element("category")
            category.attrib["id"] = _id
            category.attrib["parentId"] = "3"
            category.text = name
            objectify.deannotate(category, cleanup_namespaces=True, xsi_nil=True)
            categories.append(category)

        shop.categories = categories
        objectify.deannotate(shop, cleanup_namespaces=True, xsi_nil=True)

        return shop

    def get_offer(self, product):
        appt = objectify.Element("offer")
        appt.attrib["id"] = product.sku
        appt.count = product.quantity
        if product.quantity == 0:
            appt.attrib["type"] = "on.demand"

        appt.currencyId = "RUB"
        appt.price = product.price.now
        if self.get_old_price(product):
            appt.oldprice = self.get_old_price(product)

        for image in product.images:
            appt.addattr("picture", self.get_media_url(image))

        appt.name = product.name
        appt.description = product.description.generate(self)
        appt.vendor = product.brand.title
        appt.categoryId = self.fetch_id(product.categories[0].id)
        appt.barcode = product.barcode.code if product.barcode else 0
        appt.dimensions = (
            product.dimensions if product.dimensions else "0/0/0"
        )  # Длина, ширина, высота в упаковке.
        appt.weight = product.weight if product.weight else 0

        objectify.deannotate(appt, cleanup_namespaces=True, xsi_nil=True)
        return appt



from gdshoplib.services.notion.base import BasePage
from gdshoplib.services.notion.notion import Notion


class Block(BasePage):
    def get_source(self):
        return self.notion.get_block(self.id)

    def refresh(self):
        Notion(caching=True).get_block(self.id)
        self.initialize()

    def get_capture(self):
        capture = self[self.type].get("caption")
        return capture[0].get("plain_text") if capture else ""

    def commit(self):
        # Проитерироваться по изменениям и выполнить в Notion
        ...

    def to_json(self):
        # Вернуть товар в json
        ...



import typer

from gdshoplib.cli.crm.application import app as crm
from gdshoplib.cli.finance.application import app as finance
from gdshoplib.cli.product.application import app as product
from gdshoplib.cli.service.application import app as service
from gdshoplib.cli.temporal.application import app as temporal

app = typer.Typer()

app.add_typer(crm, name="crm")
app.add_typer(temporal, name="temporal")
app.add_typer(finance, name="finance")
app.add_typer(product, name="product")
app.add_typer(service, name="service")

if __name__ == "__main__":
    app()



from typing import Optional

from pydantic import BaseModel, ValidationError

from gdshoplib.apps.products.price import Price
from gdshoplib.services.notion.models.props import PropModel


class OnRequestTable:
    def __init__(self, blocks, /, notion, parent):
        self.notion = notion
        self.parent = parent

        self._blocks = blocks
        self._rows = []

        self.__iter = None

    def _parse_rows(self):
        result = []
        for i, block in enumerate(self._blocks):
            row = []
            for r in block["table_row"]["cells"]:
                if not r:
                    row.append(None)
                    continue
                row.append(PropModel(page={}).get_type_data(r[0]))

            if row[4] == "Ошибка":
                continue

            row_result = dict(
                block_id=block["id"],
                name=row[0] or None,
                link=row[1] or None,
                price=row[2] or None,
                quantity=row[3] or None,
            )
            try:
                result.append(OnRequestTableRow(**row_result))
            except ValidationError as e:
                row_result["error"] = str(e)
                result.append(OnRequestTableInvalidRow(**row_result))
        return result

    @property
    def rows(self):
        if not self._rows:
            self._rows = self._parse_rows()
        return self._rows

    def price(self, base_price="neitral"):
        return (
            sum(
                [
                    Price(r)[base_price] * r.quantity
                    for r in filter(
                        lambda x: isinstance(x, OnRequestTableRow), self.rows
                    )
                ]
            )
            or 0
        )

    def __bool__(self):
        return bool(self.rows)

    def __iter__(self):
        self.__iter = iter(self.rows)
        return self

    def __next__(self):
        return next(self.__iter)

    def update(self):
        for r in self.rows:
            self.notion.update_block(
                r.block_id,
                params={
                    "table_row": {
                        "cells": [
                            [self.row_object_formater(r.name)],
                            [self.row_object_formater(r.link)],
                            [self.row_object_formater(r.price)],
                            [self.row_object_formater(r.quantity)],
                            [self.row_object_formater(r.error)],
                        ]
                    }
                },
            )

    def row_object_formater(self, data):
        return {
            "type": "text",
            "text": {"content": f"{data}" if data else "", "link": None},
            "annotations": {
                "bold": False,
                "italic": False,
                "strikethrough": False,
                "underline": False,
                "code": False,
                "color": "default",
            },
            "plain_text": f"{data}" if data else "",
            "href": None,
        }


class OnRequestTableRow(BaseModel):
    block_id: str
    name: str
    link: str
    price: float
    quantity: int
    error: Optional[str]

    @property
    def price_eur(self):
        return self.price


class OnRequestTableInvalidRow(BaseModel):
    block_id: str
    name: Optional[str]
    link: Optional[str]
    price: Optional[str]
    quantity: Optional[str]
    error: str



class Storage:
    def amount(self, products, base_price="net"):
        result = 0
        for product in products:
            result += product.price[base_price] * product.quantity
        return result



import hashlib
from datetime import datetime

import ujson as json
from dateutil.parser import parse

from gdshoplib.apps.products.media import ProductMedia, S3File
from gdshoplib.core.ecosystem import Ecosystem
from gdshoplib.packages.s3v2 import File, S3v2
from gdshoplib.services.notion.page import Page


class Uploader:
    ECOSYSTEM = None

    def __init__(self, topic):
        self.topic = topic
        self.accepted_blocks = ("table_row", "image", "code")

    @property
    def ecosystem(self):
        if not self.ECOSYSTEM:
            self.ECOSYSTEM = Ecosystem()
        return self.ECOSYSTEM

    def send(self, message, parent=None):
        if message["object"] == "page":
            for k, value in message["properties"].items():
                if value["type"] == "files":
                    value = {**value, "page_id": message["id"]}

                _message = dict(
                    id=value["id"],
                    title=k,
                    parent_id=message["id"],
                    type=value["type"],
                    content=json.dumps({"content": handlers[value["type"]](value)}),
                )
                self.ecosystem.send_message(f"{self.topic}.property", data=_message)

            _message = dict(
                id=message["id"],
                parent_id=message["parent"][message["parent"]["type"]],
                url=message["url"],
                last_edited_time=handlers["last_edited_time"](message),
                created_time=handlers["created_time"](message),
                author_id=message["created_by"]["id"],
                editor_id=message["last_edited_by"]["id"],
                archived=message["archived"],
            )
            self.ecosystem.send_message(f"{self.topic}.page", data=_message)

        elif message["object"] == "block":
            if message["type"] in self.accepted_blocks:
                _message = dict(
                    id=message["id"],
                    type=message["type"],
                    content=json.dumps({"content": handlers[message["type"]](message)}),
                    has_children=message["has_children"],
                    archived=message["archived"],
                    parent_id=parent or message["parent"][message["parent"]["type"]],
                )

                self.ecosystem.send_message(f"{self.topic}.block", data=_message)
                self.ecosystem.send_message(
                    f"{self.topic}.block.{message['type']}", data=_message
                )


def file_handler(data):
    # Получить ключ файла
    _object = ProductMedia(data["id"])
    key = f'{data["type"]}.{data["id"]}.{datetime.timestamp(parse(data["last_edited_time"]))}.{_object.format}'

    # Сохранить/проверить на S3
    if not S3v2().exists(File(key=key)):
        S3v2().put(File(content=_object.content, key=key, mime=_object.mime))

    return {
        "key": key,
        "caption": handlers["text"](data[data["type"]]["caption"]),
    }


def files_handler(data):
    # Получить ключ файла
    page = Page(data["page_id"])
    result = []

    for ind, file in enumerate(data["files"]):
        _file = S3File(file["file"]["url"])
        _key = f'files.{page.id}.{data["id"]}.{ind}.{datetime.timestamp(parse(page.last_edited_time))}'
        key = f'{hashlib.md5(_key.encode("utf-8")).hexdigest()}.{_file.format}'

        if not S3v2().exists(File(key=key)):
            S3v2().put(File(content=_file.content, key=key, mime=_file.mime))

        result.append(key)
    return result


def table_row_handler(data):
    # TODO: Таблицы собрать все в 1 массив массивов
    result = {"table_id": data["parent"]["block_id"], "cells": []}
    for cell in data["table_row"]["cells"]:
        result["cells"].append(handlers["text"](cell))
    return result


def date_handler(data):
    result = {"start": None, "end": None}
    _date = data["date"]
    if not _date:
        return None

    if _date["start"]:
        result["start"] = datetime.timestamp(parse(_date["start"]))

    if _date["end"]:
        result["end"] = datetime.timestamp(parse(_date["end"]))

    return result


def code_handler(data):
    result = {
        "language": data["code"]["language"],
        "caption": handlers["text"](data["code"]["caption"]),
        "content": handlers["rich_text"](data["code"]),
    }
    return result


handlers = {}

handlers["select"] = lambda data: data["select"] and data["select"]["name"]
handlers["files"] = lambda data: files_handler(data)
handlers["image"] = lambda data: file_handler(data)
handlers["multi_select"] = lambda data: [
    select["name"] for select in data["multi_select"]
]
handlers["last_edited_time"] = lambda data: datetime.timestamp(
    parse(data["last_edited_time"])
)
handlers["created_time"] = lambda data: datetime.timestamp(parse(data["created_time"]))
handlers["date"] = date_handler
handlers["created_by"] = lambda data: data["created_by"]["id"]
handlers["last_edited_by"] = lambda data: data["last_edited_by"]["id"]
handlers["relation"] = lambda data: [r["id"] for r in data["relation"]]
handlers["status"] = lambda data: data["status"]
handlers["url"] = lambda data: data["url"]
handlers["number"] = lambda data: data["number"]
handlers["rich_text"] = lambda data: "".join(
    [text[text["type"]]["content"] for text in data["rich_text"]]
)
handlers["title"] = lambda data: "".join(
    [text[text["type"]]["content"] for text in data["title"]]
)
handlers["text"] = lambda data: "".join([text["plain_text"] for text in data])
handlers["table_row"] = table_row_handler
handlers["code"] = code_handler
handlers["checkbox"] = lambda data: data["checkbox"]
handlers["formula"] = lambda data: data["formula"][data["formula"]["type"]]
handlers["people"] = lambda data: [person["id"] for person in data["people"]]
handlers["unique_id"] = lambda data: data["unique_id"]



from .avito import AvitoManager
from .base import Platform
from .instagram import InstagramManager
from .ok import OkManager
from .sm import SberMarketManager
from .tg import TgManager
from .ula import UlaManager
from .vk import VKManager
from .yam import YandexMarketManager


def search_platform(key):
    for platform_class in Platform.__subclasses__():
        if platform_class.KEY.lower() == key.lower():
            return platform_class


__all__ = (
    "AvitoManager",
    "InstagramManager",
    "OkManager",
    "TgManager",
    "UlaManager",
    "VKManager",
    "YandexMarketManager",
    "SberMarketManager",
    "Platform",
    "search_platform",
)



import hashlib
import time
from typing import Optional

import ujson as json
from kafka import KafkaProducer
from kafka.errors import KafkaError
from loguru import logger
from pydantic import BaseModel

from gdshoplib.core.settings import NotionSettings


class Message(BaseModel):
    data: dict
    data_md5: Optional[str]

    def send(self, producer, topic):
        data = json.dumps(self.data, sort_keys=True).encode("utf-8")
        self.data_md5 = hashlib.md5(data).hexdigest()
        future = producer.send(topic, {"data_md5": self.data_md5, **self.data})
        try:
            future.get(timeout=10)
        except KafkaError as e:
            logger.exception(e)


class Ecosystem:
    _producer = None

    @property
    def producer(self):
        if not self._producer:
            self._producer = KafkaProducer(
                bootstrap_servers=NotionSettings().KAFKA_BROKER,
                value_serializer=lambda m: json.dumps(m, sort_keys=True).encode(
                    "utf-8"
                ),
                request_timeout_ms=1000,
            )
        return self._producer

    @classmethod
    def kafka(func):
        def wrap(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        return wrap

    def send_message(self, topic, /, data):
        Message(data=data).send(self.producer, topic)
        logger.info(f"{time.time()}: {topic}")



from gdshoplib.services.vk.vk import VK


class VKPage:
    def __init__(self, manager=None):
        self.manager = manager or VK()

    def set_enable(self):
        response = self.manager.request(
            "groups.enableOnline",
            params={"group_id": self.manager.settings.VK_GROUP_ID},
        )
        return bool(
            response.get("response") or response.get("error", {}).get("error_code") == 8
        )

    def set_disable(self):
        response = self.manager.request(
            "groups.disableOnline",
            params={"group_id": self.manager.settings.VK_GROUP_ID},
        )
        return bool(
            response.get("response") or response.get("error", {}).get("error_code") == 8
        )

    def get_online_status(self):
        status = self.manager.request(
            "groups.getOnlineStatus",
            params={"group_id": self.manager.settings.VK_GROUP_ID},
        )["response"].get("status")
        return status == "online"



import asyncio
import random
import string

import typer
from temporalio.client import Client
from temporalio.worker import Worker

from gdshoplib.core.settings import EcosystemSettings

app = typer.Typer()


@app.command()
def workflow():
    async def action():
        client = await Client.connect(EcosystemSettings().TEMPORAL_URL)

        # Run activity worker
        async with Worker(
            client, task_queue="say-hello-task-queue", activities=["say-hello-activity"]
        ):
            # Run the Go workflow
            workflow_id = "".join(
                random.choices(string.ascii_uppercase + string.digits, k=30)
            )
            result = await client.execute_workflow(
                "say-hello-workflow",
                "Temporal",
                id=workflow_id,
                task_queue="say-hello-task-queue",
            )

            print(result)

    asyncio.run(action())


if __name__ == "__main__":
    app()



from multiprocessing import Pool
from typing import Optional

import typer

from gdshoplib.apps.products.product import Product
from gdshoplib.services.notion.notion import Notion

app = typer.Typer()


@app.command()
def scan(
    single: bool = typer.Option(False),
    sku: Optional[str] = typer.Option(None),
):
    """Отсканировать и проставить коды"""
    if sku:
        warm_func(Product.get(sku).id)
        return

    if single:
        for page in Product.query(
            notion=Notion(caching=True),
            params={
                "filter": {
                    "and": [
                        {"property": "Штрихкод", "files": {"is_not_empty": True}},
                        {"property": "Код штрихкода", "number": {"is_empty": True}},
                    ]
                }
            },
        ):
            warm_func(page["id"])
    else:
        with Pool(3) as p:
            for page in Product.query(
                notion=Notion(caching=True),
                params={
                    "filter": {
                        "and": [
                            {
                                {
                                    "property": "Штрихкод",
                                    "files": {"is_not_empty": True},
                                },
                                {
                                    "property": "Код штрихкода",
                                    "number": {"is_empty": True},
                                },
                            },
                        ]
                    }
                },
            ):
                p.apply_async(warm_func, (page["id"],))
            p.close()
            p.join()


def warm_func(id):
    product = Product(id)
    if not product.barcode_image_field:
        return
    if not product.barcode_code_field:
        if not product.barcode:
            print(f"Штрихкод не прочитан: {product}")
            return

        product.notion.update_prop(
            product.id,
            params={"properties": {"Код штрихкода": {"number": product.barcode.code}}},
        )
        print(product)



import time
from multiprocessing import Pool
from typing import Optional

import typer

from gdshoplib.apps.products.product import Product
from gdshoplib.services.notion.database import Database

app = typer.Typer()


@app.command()
def update(
    sku: Optional[str] = typer.Option(None),
    single: bool = typer.Option(False),
    loop_iteration: Optional[int] = typer.Option(None),
):
    while True:
        if sku:
            price_update_action(Product.get(sku).id)
            return

        if single:
            for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                price_update_action(product["id"])
        else:
            with Pool(3) as p:
                for product in Database(Product.SETTINGS.PRODUCT_DB).pages():
                    p.apply_async(price_update_action, (product["id"],))
                p.close()
                p.join()

        if loop_iteration:
            time.sleep(loop_iteration)
        else:
            break


###


def price_update_action(id):
    product = Product(id)
    props_map = {
        "price_now": dict(name="Текущая Цена", value=product.price.now),
        "price_neitral": dict(name="Безубыточность", value=product.price.neitral),
        "price_current_discount": dict(
            name="Текущая Скидка", value=product.price.current_discount
        ),
        "price_gross": dict(name="Себестоимость", value=product.price.gross),
    }

    for k, v in props_map.items():
        if not product[k] and not v["value"] or product[k] == v["value"]:
            continue

        product.notion.update_prop(
            product.id, params={"properties": {v["name"]: {"number": v["value"]}}}
        )
        print(f'{product.sku}: {v["name"]} ({v["value"]})')



import re
from random import randrange

from gdshoplib.apps.products.description import ProductDescription
from gdshoplib.apps.products.media import ProductMedia
from gdshoplib.apps.products.price import Price
from gdshoplib.core.settings import ProductSettings
from gdshoplib.packages.barcode_pack import Barcode
from gdshoplib.services.notion.database import Database
from gdshoplib.services.notion.page import Page


class Product(Page):
    SETTINGS = ProductSettings()

    def __init__(self, *args, **kwargs):
        self._images = None
        self._videos = None
        self._price = None
        self._description = None
        self._kit = None
        self._badges = None

        super(Product, self).__init__(*args, **kwargs)

    @classmethod
    def get(cls, sku):
        page = [
            page
            for page in Database(cls.SETTINGS.PRODUCT_DB).pages(
                params={
                    "filter": {
                        "property": "Наш SKU",
                        "rich_text": {
                            "equals": sku,
                        },
                    }
                }
            )
        ]
        if not page:
            return
        if len(page) > 1:
            raise SKUDuplicate
        page = page[0]

        return cls(page["id"], notion=page.notion, parent=page.parent)

    @classmethod
    def query(cls, filter=None, params=None, notion=None):
        for page in Database(cls.SETTINGS.PRODUCT_DB, notion=notion).pages(
            filter=filter, params=params
        ):
            yield cls(page["id"], notion=page.notion, parent=page.parent)

    @property
    def barcode(self):
        if self.barcode_code_field:
            return Barcode(self.barcode_code_field)
        if self.barcode_image_field:
            return Barcode.read(self.barcode_image_field.image)

    @property
    def media(self):
        return [*self.images, *self.videos]

    @property
    def videos(self):
        if not self._videos:
            self._videos = list(
                ProductMedia(block.id, notion=self.notion, parent=self)
                for block in self.blocks(filter={"type": "video"})
            )
        return self._videos

    @property
    def images(self):
        if not self._images:
            self._images = list(
                ProductMedia(block.id, notion=self.notion, parent=self)
                for block in self.blocks(filter={"type": "image"})
            )
        return self._images

    @property
    def price(self):
        if not self._price:
            self._price = Price(self)
        return self._price

    @property
    def description(self):
        if not self._description:
            self._description = ProductDescription(self)
        return self._description

    @staticmethod
    def split(source):
        result = []
        if not source:
            return []

        for tag in source.split("/"):
            result.append(tag.strip())

        return result

    @property
    def tags(self):
        result = []
        for tag in self.split(self.tags_field):
            result.append(re.sub(r"[\W\s]", "", tag).lower())
        return result

    @property
    def specifications(self):
        result = []
        for spec in self.split(self.specifications_field):
            result.append(spec.capitalize())
        return result

    @property
    def notes(self):
        result = []
        for spec in self.split(self.notes_field):
            result.append(spec.capitalize())
        return result

    @property
    def kit(self):
        if not self.kit_field:
            return

        if not self._kit:
            self._kit = []
            for page in self.kit_field:
                if self.id != page.id:
                    self._kit.append(
                        Product(page.id, notion=self.notion, parent=self.parent)
                    )
        return self._kit

    @property
    def platforms(self):
        # Получить платформы из фида и вернуть объект класса <Platform>
        ...

    @property
    def badges(self):
        if not self.badge_field:
            return

        if not self._badges:
            self._badges = []
            for page in self.badge_field:
                if self.id != page.id:
                    self._badges.append(
                        Page(page.id, notion=self.notion, parent=self.parent)
                    )
        return self._badges

    def available(self):
        return bool(
            self.quantity
            and self.status_publication == "Публикация"
            and self.status_description == "Готово"
        )

    def generate_sku(self):
        # Сгенерировать SKU на основе продукта
        # Категория.Бренд.Цена_покупки.месяц_добавления.год_добавления.случайные_4_числа

        sku = (
            f"{self.brand.title.upper() if self.brand else ''}"
            f"{int(self.price.eur)}"
            f"{randrange(1111, 9999)}"
        )

        return re.sub(r"\W", "", sku)

    def __str__(self) -> str:
        return f"{self.__class__}: {self.sku}"

    def __repr__(self) -> str:
        return f"{self.__class__}: {self.sku}"

    def warm(self):
        self.price.now
        self.kit
        self.notes
        self.specifications
        self.tags
        self.media
        self.description
        self.brand.title
        self.description.warm_description_blocks()


class SKUDuplicate(Exception):
    # Если найден товар с одинаковым SKU
    ...



from datetime import datetime

import typer

from gdshoplib.core.ecosystem import Ecosystem
from gdshoplib.services.avito.avito import Avito
from gdshoplib.services.vk.market import VKMarket
from gdshoplib.services.vk.page import VKPage
from gdshoplib.services.vk.stats import VKStats
from gdshoplib.services.vk.vk import VK

app = typer.Typer()


@app.command()
def access_code(code=None):
    if not code:
        VK().get_oauth_code()
        code = typer.prompt("Код")

    print(VK().get_access_token(code))


@app.command()
def online(active: bool = True):
    if active:
        VKPage().set_enable()
    else:
        VKPage().set_disable()


@app.command()
def health():
    assert VKMarket().list(), "Запрос в VK не выполняется"
    print("OK")


@app.command()
def upload():
    ecosystem = Ecosystem()
    for message in [
        *VKStats().community(),
        *VKStats().products(),
        *VKStats().members(),
    ]:
        ecosystem.send_message("vk", data=message, message_type="stats")


if __name__ == "__main__":
    app()



import typer

from gdshoplib.apps.finance.storage import Storage
from gdshoplib.apps.products.product import Product

app = typer.Typer()


@app.command()
def amount(base_price="now"):
    params = {"filter": {"property": "Кол-во", "number": {"greater_than": 0}}}
    print(
        f"{base_price}: {Storage().amount(Product.query(params=params), base_price=base_price)}"
    )



from typing import Optional

import typer

from gdshoplib.apps.products.product import Product
from gdshoplib.services.notion.notion import Notion

app = typer.Typer()


@app.command()
def sku_set():
    for page in Product.query(
        notion=Notion(caching=True),
        params={
            "filter": {
                "and": [
                    {"property": "Наш SKU", "rich_text": {"is_empty": True}},
                    {"property": "Цена (eur)", "number": {"is_not_empty": True}},
                ]
            }
        },
    ):
        sku = page.generate_sku()
        while not Product.query(filter={"sku": sku}):
            sku = page.generate_sku()

        page.notion.update_prop(
            page.id, params={"properties": {"Наш SKU": [{"text": {"content": sku}}]}}
        )
        print(Product(page.id).sku)


@app.command()
def update_sku_naming(sku: Optional[str] = typer.Option(None)):
    sku_filter = {"equals": sku} if sku else {"is_not_empty": True}
    for page in Product.query(
        notion=Notion(caching=True),
        params={
            "filter": {
                "and": [
                    {"property": "Наш SKU", "rich_text": sku_filter},
                    {"property": "Описание", "status": {"equals": "Готово"}},
                ]
            }
        },
    ):
        title = f"{page.name.strip()}:{page.sku}"
        if page.title != title:
            if not page.original_name:
                page.notion.update_prop(
                    page.id,
                    params={
                        "properties": {
                            "Оригинальное название": [{"text": {"content": page.title}}]
                        }
                    },
                )
            page.notion.update_prop(
                page.id,
                params={"properties": {"title": [{"text": {"content": title}}]}},
            )
            print(title)



from typing import List, Optional

from dateutil.parser import parse
from pydantic import BaseModel


class PropModel(BaseModel):
    id: Optional[str]
    key: Optional[str]
    name: Optional[str]

    page: dict
    fields: Optional[List[dict]]

    def get_data(self):
        field = None
        page_field = None

        if not self.fields:
            raise KeyError

        for field_context in self.fields:
            _page_field = self.page["properties"].get(field_context["name"])
            if _page_field and _page_field["id"] == field_context["id"]:
                field = field_context
                page_field = _page_field

        if not page_field:
            return

        data = self.get_type_data(page_field)
        handler = field.get("handler")

        if handler and data:
            return handler(data)
        return data

    def get_type_data(self, page_field):
        return self.properties_type_parse_map.get(page_field["type"])(page_field)

    @property
    def properties_type_parse_map(self):
        return {
            "rich_text": lambda data: " ".join(
                [t.get("plain_text", "") for t in data["rich_text"]]
            )
            or "",
            "text": lambda data: data["plain_text"] or "",
            "number": lambda data: data["number"] or 0,
            "title": lambda data: data["title"][0]["text"]["content"],
            "select": lambda data: data.get("select").get("name")
            if data.get("select")
            else None,
            "multi_select": lambda data: data,
            "status": lambda data: data["status"]["name"],
            "date": lambda data: data["date"],
            "formula": lambda data: data["formula"]["number"],
            "relation": lambda data: data["relation"],
            "rollup": lambda data: data,
            "people": lambda data: data,
            "files": lambda data: data["files"],
            "checkbox": lambda data: data["checkbox"],
            "url": lambda data: data["url"],
            "email": lambda data: data,
            "phone_number": lambda data: data,
            "created_time": lambda data: parse(data["created_time"]),
            "created_by": lambda data: str(data["created_by"]),
            "last_edited_time": lambda data: parse(data["last_edited_time"]),
            "last_edited_by": lambda data: str(data["last_edited_by"]),
            "image": lambda data: data["image"]["file"]["url"],
            "video": lambda data: data["video"]["file"]["url"],
        }



import time

from gdshoplib.apps.crm.on_request import (OnRequestTable,
                                           OnRequestTableInvalidRow)
from gdshoplib.apps.products.price import Price
from gdshoplib.apps.products.product import Product
from gdshoplib.core.settings import CRMSettings
from gdshoplib.services.notion.database import Database
from gdshoplib.services.notion.notion import Notion
from gdshoplib.services.notion.page import Page


class Order(Page):
    SETTINGS = CRMSettings()

    def __init__(self, *args, **kwargs):
        self._on_request = None
        self._products = None

        super(Order, self).__init__(*args, **kwargs)

    def __len__(self):
        return len(self.products) + sum([row.quantity for row in self.on_request])

    @staticmethod
    def list_filter(filter_type="raw"):
        if filter_type == "raw":
            return {
                "filter": {
                    "or": [
                        {"property": "Status", "status": {"equals": "Неразобранное"}},
                    ]
                }
            }
        elif filter_type == "complited_shop":
            return {
                "filter": {
                    "and": [
                        {"property": "Status", "status": {"equals": "Доставлено"}},
                        {"property": "Тип заказа", "select": {"equals": "Магазин"}},
                    ]
                }
            }
        elif filter_type == "complited_agent":
            return {
                "filter": {
                    "and": [
                        {"property": "Status", "status": {"equals": "Доставлено"}},
                        {"property": "Тип заказа", "select": {"equals": "Agent"}},
                    ]
                }
            }
        elif filter_type == "complited":
            return {
                "filter": {
                    "or": [
                        {
                            "property": "Status",
                            "status": {"does_not_equal": "Доставлено"},
                        },
                        {"property": "Status", "status": {"does_not_equal": "Отказ"}},
                        {
                            "property": "Status",
                            "status": {"does_not_equal": "Оформление"},
                        },
                        {"property": "Status", "status": {"does_not_equal": "Оплата"}},
                    ]
                }
            }

    @classmethod
    def get(cls, id):
        page = [
            page
            for page in Database(cls.SETTINGS.CRM_DB).pages(
                params={
                    "filter": {
                        "property": "ID заказа",
                        "rich_text": {
                            "equals": id,
                        },
                    }
                }
            )
        ]
        if not page:
            return
        if len(page) > 1:
            raise OrderIDDuplicate

        page = page[0]

        return cls(page["id"], notion=page.notion, parent=page.parent)

    @classmethod
    def query(cls, filter=None, params=None, notion=None):
        for page in Database(cls.SETTINGS.CRM_DB, notion=notion).pages(
            filter=filter, params=params
        ):
            yield cls(page["id"], notion=page.notion, parent=page.parent)

    @property
    def on_request(self):
        if not self._on_request:
            self._on_request = OnRequestTable(
                [block for block in self.blocks(filter={"type": "table_row"})],
                notion=self.notion,
                parent=self,
            )

        return self._on_request

    @property
    def products(self):
        if not self._products:
            self._products = [Product(page.id) for page in self.products_field]
        return self._products

    def requested_price(self, product):
        return product.price.profit - product.price.profit * (self.discount * 0.01)

    @property
    def delivery(self):
        return

    @property
    def pay(self):
        return

    @property
    def discount_sum(self):
        return round(self.full_price - self.price, 2)

    @property
    def max_discount(self):
        # Расчет максимальной скидки
        if self.order_type != "Магазин":
            return 0

        now = self.on_request.price()
        for product in self.products:
            now += product.price.now
        lowest = self.on_request.price()
        for product in self.products:
            lowest += product.price.neitral

        result = now - lowest
        if not result:
            return 0
        return round((result) / (now * 0.01))

    def get_promocode(self):
        if not self.promocodes or self.order_type != "Магазин":
            return None
        self.promocodes.sort(key=lambda x: x.discount)
        return self.promocodes[0]

    @property
    def price(self):
        result = self.on_request.price()
        for product in self.products:
            if self.order_type == "Агент":
                result += product.price.neitral
            else:
                promocode = self.get_promocode()
                if promocode and product.price.current_discount < promocode.discount:
                    _price = (
                        product.price.profit
                        if product.price.profit > product.price.now
                        else product.price.now
                    )
                    result += _price - _price * (promocode.discount * 0.01)
                else:
                    result += product.price.now

        return round(result - result * (self.discount * 0.01))

    @property
    def full_price(self):
        result = self.on_request.price()
        for product in self.products:
            result += (
                product.price.profit
                if product.price.profit > product.price.now
                else product.price.now
            )

        return result

    @property
    def profit(self):
        base_price = self.on_request.price("gross")
        for product in self.products:
            base_price += product.price.gross
        return self.price - base_price

    def generate_id(self):
        return f"{self.platform.key.lower()}.{int(time.time())}"

    def description_on_request_product(self, row):
        quantity = f"({row.quantity} шт.)"
        return f"{row.name} {quantity if row.quantity > 1 else ''}: {Price(row)['neitral'] * row.quantity} ₽"

    def description_basic_product(self, product):
        _price = (
            product.price.profit
            if product.price.profit > product.price.now
            else product.price.now
        )
        return f"{product.name}: {_price} ₽"

    @property
    def description(self):
        result = ["Ваш заказ\n"]
        counter = 1

        for row in self.on_request:
            if isinstance(row, OnRequestTableInvalidRow):
                continue
            result.append(f"{counter}. {self.description_on_request_product(row)} ₽")
            counter += 1

        for product in self.products:
            result.append(f"{counter}. {self.description_basic_product(product)} ₽")
            counter += 1

        if self.delivery or self.discount_sum:
            result.append(f"\nТовары ({len(self)}): {self.full_price}")

        if self.get_promocode():
            promocode = self.get_promocode()
            result.append(f"Промокод: {promocode.title} (-{promocode.discount}%)")

        if self.discount_sum:
            result.append(f"Скидка: {self.discount_sum} ₽")

        if self.delivery:
            result.append(f"Стоимость доставки: {self.delivery.price} ₽")

        result.append(f"\nОбщая стоимость: {self.price} ₽")

        return "\n".join(result)

    def set_id(self):
        # Установить ID
        if not self.platform:
            return

        self.notion.update_prop(
            self.id,
            params={
                "properties": {"ID заказа": [{"text": {"content": self.generate_id()}}]}
            },
        )

    def set_max_discount(self):
        # Установка максимальной скидки
        self.notion.update_prop(
            self.id,
            params={"properties": {"max Скидка": {"number": self.max_discount or 0}}},
        )

    def set_price_description(self):
        # Установка описания расчетов цены
        self.notion.update_prop(
            self.id,
            params={
                "properties": {
                    "Расчеты": [{"text": {"content": self.description or ""}}]
                }
            },
        )

    def set_price(self):
        # Установка итоговой цены
        self.notion.update_prop(
            self.id,
            params={"properties": {"Итого": {"number": self.price}}},
        )

    def set_profit(self):
        # Установка итоговой цены
        self.notion.update_prop(
            self.id,
            params={"properties": {"Прибыль": {"number": self.profit}}},
        )

    def set_updated(self):
        self.notion.update_prop(
            self.id,
            params={
                "properties": {
                    "Прибыль": {"number": self.profit},
                    "Итого": {"number": self.price},
                    "Расчеты": [{"text": {"content": self.description or ""}}],
                    "max Скидка": {"number": self.max_discount or 0},
                }
            },
        )

    def load_tasks(self):
        # Загрузка задач в CRM из источников
        ...

    @classmethod
    def update(cls):
        # Обновление карточки Order
        for order in cls.query(notion=Notion(caching=True), params=cls.list_filter()):
            order.on_request.update()
            order.set_price_description()
            order.set_profit()
            order.set_price()
            order.set_max_discount()
            if not order.pk:
                order.set_id()

    def notification(self):
        # Отправка уведомлений
        ...


class OrderIDDuplicate(Exception):
    ...



import datetime

from lxml import etree, objectify

from gdshoplib.apps.platforms.yam import YandexMarketManager
from gdshoplib.packages.feed import Feed


class SberMarketManager(YandexMarketManager, Feed):
    KEY = "SM"

    def get_root(self):
        root = etree.Element("yml_catalog")
        objectify.deannotate(root, cleanup_namespaces=True, xsi_nil=True)
        root.attrib["date"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M")
        return root



from datetime import datetime

import typer

from gdshoplib.core.ecosystem import Ecosystem
from gdshoplib.services.avito.avito import Avito

app = typer.Typer()


@app.command()
def access_code():
    print(Avito().get_access_token())


@app.command()
def upload():
    ecosystem = Ecosystem()
    for message in Avito().get_statistic():
        ecosystem.send_message("avito", data=message, message_type="stats")


# if __name__ == "__main__":
#     app()



import datetime
import re
import time

import orjson as json
import redis

from gdshoplib.core.settings import CacheSettings


class BaseCache:
    @classmethod
    def get_class(cls, key):
        for _class in cls.__subclasses__():
            if _class.__name__ == key:
                return _class
        return DumpCache


class KeyDBCache(BaseCache):
    CONNECT = None

    def __init__(self, namespace=None):
        self.settings = CacheSettings()
        self.namespace = namespace or self.settings.CACHE_HSTORE
        self.system_namespace = self.settings.CACHE_SYSTEM_HSTORE
        self._update_time = None

    def subscribe(self, topic):
        p = self.connect.pubsub()
        try:
            p.subscribe(topic)
            while True:
                message = p.get_message()
                if message:
                    yield json.loads(str(message["data"]))
                time.sleep(0.001)
        finally:
            p.close()

    def publish(self, topic, value):
        self.connect.publish(topic, json.dumps(value))

    @staticmethod
    def update_time_key(id, /, type):
        return f"{type}:{id}_update_time"

    def set_update_start(self):
        self._update_time = datetime.datetime.now(datetime.timezone.utc)

    def clean_update(self):
        self._update_time = None

    @property
    def update_time(self):
        if self._update_time:
            return self._update_time
        return datetime.datetime.now(datetime.timezone.utc)

    def get_update_time(self, id, /, type):
        value = self.connect.hget(
            self.settings.CACHE_SYSTEM_HSTORE, self.update_time_key(id, type)
        )
        if value:
            return json.loads(value)

    def set_update_time(self, id, /, type):
        content = json.dumps(self.update_time)
        self.connect.hset(
            self.settings.CACHE_SYSTEM_HSTORE, self.update_time_key(id, type), content
        )

    def clean_system_cache(self):
        self.connect.delete(self.settings.CACHE_SYSTEM_HSTORE)

    @property
    def connect(self):
        if not self.CONNECT:
            self.CONNECT = redis.Redis.from_url(self.settings.CACHE_DSN)
        return self.CONNECT

    def get(self, key, default=None):
        return self[key] or default

    def __getitem__(self, key):
        value = self.connect.hget(self.settings.CACHE_HSTORE, key)
        if value:
            return json.loads(value)

    def __setitem__(self, key, value):
        content = value if isinstance(value, bytes) else json.dumps(value)
        self.connect.hset(self.settings.CACHE_HSTORE, key, content)

    def count(self):
        return self.connect.hlen(self.settings.CACHE_HSTORE)

    def delete(self, key):
        self.connect.hdel(self.settings.CACHE_HSTORE, key)

    def exists(self, key):
        return self.connect.hexists(self.settings.CACHE_HSTORE, key)

    def search(self, pattern):
        result = []
        for k in self.connect.hkeys(self.settings.CACHE_HSTORE):
            if re.match(pattern, k.decode("utf-8")):
                result.append(k.decode("utf-8"))
        return result

    def clean(self, pattern=r"*"):
        for key in self.search(pattern):
            self.delete(key)


class DumpCache(BaseCache):
    CACHE = {}

    def __init__(self, *args, dsn, cache_period):
        self.cache_period = cache_period

    def get(self, *args, **kwargs):
        return self.CACHE.get(*args, **kwargs)

    def __getitem__(self, key):
        cached = self.CACHE.get(key)
        if cached and cached["time"] > datetime.datetime.now():
            return self.CACHE[key]["data"]

    def __setitem__(self, key, value):
        self.CACHE[key] = {
            "data": value,
            "time": datetime.datetime.now()
            + datetime.timedelta(seconds=self.cache_period),
        }
